'''
Created on 18.09.2019

@author: Samuel
'''
import sys
import os

import numpy as np
import matplotlib
if sys.platform != 'win32':
    # if executed on a Windows server. Comment out this line, if you are working
    # on a desktop computer that is not Windows.
    matplotlib.use('Agg')

try:
    from test_ci_rvm import *
except ModuleNotFoundError:
    from .test_ci_rvm import *

if __name__ == '__main__':
    try:
        # if vemomoto_core_tools is installed, you can call the module with an 
        # argument specifying a log file to which the output is written in addition
        # to the console.
        from vemomoto_core.tools.tee import Tee
        if len(sys.argv) > 1:
            teeObject = Tee(sys.argv[1])
    except ModuleNotFoundError:
        pass


if __name__ == '__main__':
    #os.chdir(os.path.join("test_CI", "Histone_H1"))
    
    methods = ["mixed_min", "constrained_max", "binsearch", "bisection", "RVM"]
    methods = ["Wald", "RVM", "RVM_psI", "bisection", "mixed_min", "constrained_max", "binsearch", "VM",  "gridsearch"]
    methods = ["RVM_psI"]
    #methods = ["bisection"]
    benchmark(methods, 200, dataN=50, mode="11cx") 
    sys.exit()
    benchmark(methods, 200, dataN=10000, mode="11") 
    test_LogRegress(methods)
    test_LogRegress_pred()
    test_dynamical_system()
    test_H14()
    print(os.getcwd())



'''
Created on 03.10.2019

@author: Samuel
'''
from itertools import cycle
import re

import numpy as np
import matplotlib.pyplot as plt

COMPARE_PSEUDOINVERSE = True
FOLDER = "plots\\"
METHODS = ['Wald', 'RVM', 'RVM_psI', 'bisection', 'mixed_min', 'constrained_max', 'binsearch', 'VM', 'gridsearch']
#PLOTORDER = [ 7   ,    0 ,      3     ,       5     ,          6      ,        4  ,      1 ,   2       ]
PLOTORDER = [ 7   ,    0 ,     0.5,        3     ,       5     ,          6      ,        4  ,      1 ,   2       ]
METHODNAMES = ['Wald', 'RVM', 'RVM-MPI', 'Bisection', 'Neale-Miller', 'Constr. max.', 'Binary search', 'VM', 'Grid search']



DATA = [
    ["3", 500,
"""
Wald           : success=0.00000, error=329.5709, errorReduced=  3.8879, errorTot= 29.6541, largeErrors=0.16750, largeErrorsTot=0.16750, evals=     0.0, evalsTot=     0.0
RVM            : success=0.98583, error=  1.1031, errorReduced=  0.0045, errorTot=  0.0050, largeErrors=0.00083, largeErrorsTot=0.00083, evals=   774.3, evalsTot=  1525.9
RVM_psI        : success=0.98667, error=  1.1019, errorReduced=  0.0045, errorTot=  0.0052, largeErrors=0.00083, largeErrorsTot=0.00083, evals=   673.6, evalsTot=  1451.4
bisection      : success=0.77750, error=  0.2663, errorReduced=  0.0691, errorTot= 52.2622, largeErrors=0.03833, largeErrorsTot=0.06250, evals=  1532.6, evalsTot=  7350.2
mixed_min      : success=0.41500, error=  1.6748, errorReduced=  0.1446, errorTot=106.9708, largeErrors=0.00083, largeErrorsTot=0.09167, evals=   739.8, evalsTot=  1374.4
constrained_max: success=0.54167, error= 11.7703, errorReduced=  0.1839, errorTot=  5.3583, largeErrors=0.01750, largeErrorsTot=0.02083, evals=   581.3, evalsTot=  1054.8
binsearch      : success=0.80500, error=  0.0002, errorReduced=  0.0002, errorTot= 23.3635, largeErrors=0.00000, largeErrorsTot=0.01250, evals=  2194.5, evalsTot= 13465.2
VM             : success=0.08583, error= 31.8943, errorReduced=  0.2506, errorTot=  0.7301, largeErrors=0.00583, largeErrorsTot=0.00583, evals=   552.0, evalsTot=   317.0
gridsearch     : success=0.86500, error=168.1895, errorReduced=  0.0005, errorTot=204.6465, largeErrors=0.05000, largeErrorsTot=0.10167, evals=  8649.8, evalsTot=  9596.3
"""],
    ["3", 1000,
"""
Wald           : success=0.00000, error=275.0677, errorReduced=  4.5724, errorTot= 22.1636, largeErrors=0.18417, largeErrorsTot=0.18417, evals=     0.0, evalsTot=     0.0
RVM            : success=0.92583, error=  0.0002, errorReduced=  0.0002, errorTot=  0.0002, largeErrors=0.00000, largeErrorsTot=0.00000, evals=   404.0, evalsTot=  1350.9
RVM_psI        : success=0.91083, error=  0.0002, errorReduced=  0.0002, errorTot=  0.0002, largeErrors=0.00000, largeErrorsTot=0.00000, evals=   313.8, evalsTot=  1285.7
bisection      : success=0.78417, error=  0.2019, errorReduced=  0.0578, errorTot=  6.5979, largeErrors=0.05500, largeErrorsTot=0.06250, evals=  1335.7, evalsTot=  5618.0
mixed_min      : success=0.44333, error=  0.1066, errorReduced=  0.1066, errorTot=167.4518, largeErrors=0.00000, largeErrorsTot=0.14500, evals=   776.3, evalsTot=  1216.5
constrained_max: success=0.62000, error= 11.6432, errorReduced=  0.1671, errorTot= 10.7497, largeErrors=0.01917, largeErrorsTot=0.02750, evals=   618.4, evalsTot=  1084.8
binsearch      : success=0.87833, error=  0.0002, errorReduced=  0.0002, errorTot=  4.4954, largeErrors=0.00000, largeErrorsTot=0.00333, evals=  1987.1, evalsTot= 10605.7
VM             : success=0.22583, error= 12.5513, errorReduced=  0.1346, errorTot=  2.7998, largeErrors=0.00750, largeErrorsTot=0.00750, evals=   437.6, evalsTot=   339.6
gridsearch     : success=0.89417, error= 31.0430, errorReduced=  0.0006, errorTot= 45.0764, largeErrors=0.01250, largeErrorsTot=0.05500, evals=  6909.5, evalsTot=  7986.1
"""],
    ["3", 3000,
"""
Wald           : success=0.00000, error= 59.3576, errorReduced=  4.6430, errorTot= 18.8396, largeErrors=0.11083, largeErrorsTot=0.11083, evals=     0.0, evalsTot=     0.0
RVM            : success=0.98417, error=  0.0042, errorReduced=  0.0042, errorTot=  0.0042, largeErrors=0.00000, largeErrorsTot=0.00000, evals=   244.7, evalsTot=   390.5
RVM_psI        : success=0.98417, error=  0.0042, errorReduced=  0.0042, errorTot=  0.0042, largeErrors=0.00000, largeErrorsTot=0.00000, evals=   244.9, evalsTot=   384.9
bisection      : success=0.91417, error=  0.3843, errorReduced=  0.0799, errorTot=  0.3784, largeErrors=0.00833, largeErrorsTot=0.00833, evals=   402.0, evalsTot=  1351.4
mixed_min      : success=0.59250, error=  0.3191, errorReduced=  0.2959, errorTot=221.8844, largeErrors=0.00167, largeErrorsTot=0.22917, evals=   757.0, evalsTot=  1057.9
constrained_max: success=0.79667, error=  2.1308, errorReduced=  0.2337, errorTot= 11.5054, largeErrors=0.01250, largeErrorsTot=0.02417, evals=   649.1, evalsTot=   960.0
binsearch      : success=0.98167, error=  0.0006, errorReduced=  0.0006, errorTot=  0.0006, largeErrors=0.00000, largeErrorsTot=0.00000, evals=  1207.5, evalsTot=  2418.7
VM             : success=0.62083, error=  1.5650, errorReduced=  0.0093, errorTot=  1.9579, largeErrors=0.01417, largeErrorsTot=0.03250, evals=   303.0, evalsTot=   432.2
gridsearch     : success=0.94667, error=  0.0007, errorReduced=  0.0007, errorTot=  5.5064, largeErrors=0.00167, largeErrorsTot=0.04000, evals=  3144.7, evalsTot=  3619.0
"""],
    ["3", 10000,
"""
Wald           : success=0.00000, error=  3.7752, errorReduced=  3.4170, errorTot=  3.7752, largeErrors=0.01083, largeErrorsTot=0.01083, evals=     0.0, evalsTot=     0.0
RVM            : success=1.00000, error=  0.0001, errorReduced=  0.0001, errorTot=  0.0001, largeErrors=0.00000, largeErrorsTot=0.00000, evals=   169.5, evalsTot=   169.5
RVM_psI        : success=1.00000, error=  0.0001, errorReduced=  0.0001, errorTot=  0.0001, largeErrors=0.00000, largeErrorsTot=0.00000, evals=   169.5, evalsTot=   169.5
bisection      : success=1.00000, error=  0.0103, errorReduced=  0.0103, errorTot=  0.0103, largeErrors=0.00000, largeErrorsTot=0.00000, evals=   243.1, evalsTot=   243.1
mixed_min      : success=0.79917, error=  0.4284, errorReduced=  0.3742, errorTot= 10.0984, largeErrors=0.00500, largeErrorsTot=0.01500, evals=   751.5, evalsTot=   762.2
constrained_max: success=0.85833, error=  0.1481, errorReduced=  0.1481, errorTot=  3.4722, largeErrors=0.00000, largeErrorsTot=0.00333, evals=   554.8, evalsTot=   603.3
binsearch      : success=1.00000, error=  0.0002, errorReduced=  0.0002, errorTot=  0.0002, largeErrors=0.00000, largeErrorsTot=0.00000, evals=   912.0, evalsTot=   912.0
VM             : success=0.87750, error=  0.0893, errorReduced=  0.0519, errorTot=  0.4128, largeErrors=0.00083, largeErrorsTot=0.00083, evals=   237.8, evalsTot=   580.7
gridsearch     : success=1.00000, error=  0.0007, errorReduced=  0.0007, errorTot=  0.0007, largeErrors=0.00000, largeErrorsTot=0.00000, evals=  1788.4, evalsTot=  1788.4
"""],
    ["11cx", 50,
"""
Wald           : success=0.41318, error= 94.9196, errorReduced=  0.4698, errorTot=177.6506, largeErrors=0.08727, largeErrorsTot=0.08727, evals=     0.0, evalsTot=     0.0
RVM            : success=0.91932, error=  0.0003, errorReduced=  0.0003, errorTot= 35.3982, largeErrors=0.00750, largeErrorsTot=0.06273, evals=  2273.4, evalsTot= 19212.6
RVM_psI        : success=0.97386, error=  0.0001, errorReduced=  0.0001, errorTot= 13.5498, largeErrors=0.01750, largeErrorsTot=0.02364, evals=  1649.5, evalsTot= 16011.4
bisection      : success=0.39114, error=  2.6273, errorReduced=  0.0016, errorTot=112.1667, largeErrors=0.35432, largeErrorsTot=0.45523, evals=  1905.1, evalsTot= 32364.8
mixed_min      : success=0.24386, error=  6.6671, errorReduced=  0.5488, errorTot= 15.7724, largeErrors=0.02841, largeErrorsTot=0.03432, evals=  2661.1, evalsTot=  4685.6
constrained_max: success=0.46227, error=  5.6586, errorReduced=  0.0058, errorTot=  8.4952, largeErrors=0.00341, largeErrorsTot=0.01000, evals=  1203.2, evalsTot=  1356.3
binsearch      : success=0.52500, error=  0.0003, errorReduced=  0.0003, errorTot= 32.2177, largeErrors=0.00000, largeErrorsTot=0.01318, evals=  7902.6, evalsTot= 39368.6
VM             : success=0.35045, error=  0.0711, errorReduced=  0.0001, errorTot=  0.0711, largeErrors=0.00114, largeErrorsTot=0.00114, evals=   850.0, evalsTot=   511.8
gridsearch     : success=0.49500, error= 24.3749, errorReduced=  0.0063, errorTot=120.1377, largeErrors=0.42909, largeErrorsTot=0.49364, evals= 37619.1, evalsTot= 64827.9
"""],
    ["11cx", 100,
"""
Wald           : success=0.21000, error= 13.6389, errorReduced=  0.2674, errorTot= 13.9721, largeErrors=0.01545, largeErrorsTot=0.01545, evals=     0.0, evalsTot=     0.0
RVM            : success=0.98477, error=  0.0001, errorReduced=  0.0001, errorTot=  1.3701, largeErrors=0.00068, largeErrorsTot=0.01409, evals=   664.2, evalsTot=  2195.0
RVM_psI        : success=0.99841, error=  0.0001, errorReduced=  0.0001, errorTot=  0.0176, largeErrors=0.00091, largeErrorsTot=0.00136, evals=   662.4, evalsTot=  1693.7
bisection      : success=0.94750, error=  0.2426, errorReduced=  0.0002, errorTot=  6.5341, largeErrors=0.02523, largeErrorsTot=0.04227, evals=   907.2, evalsTot=  3701.5
mixed_min      : success=0.76455, error=  0.8103, errorReduced=  0.1259, errorTot=  0.9261, largeErrors=0.00636, largeErrorsTot=0.00682, evals=  1536.9, evalsTot=  1779.9
constrained_max: success=0.95227, error=  0.3252, errorReduced=  0.0047, errorTot=  0.2704, largeErrors=0.00068, largeErrorsTot=0.00091, evals=   760.7, evalsTot=   797.6
binsearch      : success=0.96750, error=  0.0003, errorReduced=  0.0003, errorTot=  0.6173, largeErrors=0.00000, largeErrorsTot=0.00091, evals=  4369.1, evalsTot=  7323.5
VM             : success=0.94500, error=  0.0001, errorReduced=  0.0001, errorTot=  0.0001, largeErrors=0.00000, largeErrorsTot=0.00000, evals=   739.3, evalsTot=   716.6
gridsearch     : success=0.94977, error=  2.3037, errorReduced=  0.0009, errorTot=  7.8223, largeErrors=0.03250, largeErrorsTot=0.04795, evals= 15137.6, evalsTot= 19290.7
"""],
    ["11cx", 300,
"""
Wald           : success=0.81295, error=  0.0244, errorReduced=  0.0244, errorTot=  0.0244, largeErrors=0.00000, largeErrorsTot=0.00000, evals=     0.0, evalsTot=     0.0
RVM            : success=1.00000, error=  0.0000, errorReduced=  0.0000, errorTot=  0.0000, largeErrors=0.00000, largeErrorsTot=0.00000, evals=   451.0, evalsTot=   451.0
RVM_psI        : success=1.00000, error=  0.0000, errorReduced=  0.0000, errorTot=  0.0000, largeErrors=0.00000, largeErrorsTot=0.00000, evals=   451.0, evalsTot=   451.0
bisection      : success=1.00000, error=  0.0000, errorReduced=  0.0000, errorTot=  0.0000, largeErrors=0.00000, largeErrorsTot=0.00000, evals=   904.6, evalsTot=   904.6
mixed_min      : success=0.95023, error=  0.0077, errorReduced=  0.0077, errorTot=  0.0077, largeErrors=0.00000, largeErrorsTot=0.00000, evals=  1087.1, evalsTot=  1087.1
constrained_max: success=0.98886, error=  0.0018, errorReduced=  0.0018, errorTot=  0.0018, largeErrors=0.00000, largeErrorsTot=0.00000, evals=   607.9, evalsTot=   607.9
binsearch      : success=1.00000, error=  0.0005, errorReduced=  0.0005, errorTot=  0.0005, largeErrors=0.00000, largeErrorsTot=0.00000, evals=  3258.9, evalsTot=  3258.9
VM             : success=1.00000, error=  0.0000, errorReduced=  0.0000, errorTot=  0.0000, largeErrors=0.00000, largeErrorsTot=0.00000, evals=   602.8, evalsTot=   602.8
gridsearch     : success=0.99864, error=  0.0008, errorReduced=  0.0008, errorTot=  0.0008, largeErrors=0.00000, largeErrorsTot=0.00000, evals= 10721.5, evalsTot= 10721.5
"""],
    ["11cx", 1000,
"""
Wald           : success=0.98682, error=  0.0057, errorReduced=  0.0057, errorTot=  0.0057, largeErrors=0.00000, largeErrorsTot=0.00000, evals=     0.0, evalsTot=     0.0
RVM            : success=1.00000, error=  0.0000, errorReduced=  0.0000, errorTot=  0.0000, largeErrors=0.00000, largeErrorsTot=0.00000, evals=   451.0, evalsTot=   451.0
RVM_psI        : success=1.00000, error=  0.0000, errorReduced=  0.0000, errorTot=  0.0000, largeErrors=0.00000, largeErrorsTot=0.00000, evals=   451.0, evalsTot=   451.0
bisection      : success=1.00000, error=  0.0000, errorReduced=  0.0000, errorTot=  0.0000, largeErrors=0.00000, largeErrorsTot=0.00000, evals=   942.0, evalsTot=   942.0
mixed_min      : success=0.99614, error=  0.0019, errorReduced=  0.0019, errorTot=  0.0019, largeErrors=0.00000, largeErrorsTot=0.00000, evals=   928.2, evalsTot=   928.2
constrained_max: success=0.97818, error=  0.0030, errorReduced=  0.0030, errorTot=  0.0030, largeErrors=0.00000, largeErrorsTot=0.00000, evals=   663.0, evalsTot=   663.0
binsearch      : success=1.00000, error=  0.0005, errorReduced=  0.0005, errorTot=  0.0005, largeErrors=0.00000, largeErrorsTot=0.00000, evals=  3298.6, evalsTot=  3298.6
VM             : success=1.00000, error=  0.0000, errorReduced=  0.0000, errorTot=  0.0000, largeErrors=0.00000, largeErrorsTot=0.00000, evals=   597.0, evalsTot=   597.0
gridsearch     : success=0.99864, error=  0.0008, errorReduced=  0.0008, errorTot=  0.0008, largeErrors=0.00000, largeErrorsTot=0.00000, evals=  9786.2, evalsTot=  9786.2
"""],
    ["11", 500,
"""
Wald           : success=0.02318, error=303.4153, errorReduced=  0.8097, errorTot= 61.7644, largeErrors=0.05477, largeErrorsTot=0.05477, evals=     0.0, evalsTot=     0.0
RVM            : success=0.89659, error=  4.8438, errorReduced=  0.0473, errorTot=  6.2265, largeErrors=0.00932, largeErrorsTot=0.01455, evals=  3961.5, evalsTot= 11800.9
RVM_psI        : success=0.82114, error=  4.0107, errorReduced=  0.0363, errorTot= 50.1538, largeErrors=0.02545, largeErrorsTot=0.04182, evals=  4656.8, evalsTot= 14374.8
bisection      : success=0.70886, error= 83.8969, errorReduced=  0.0592, errorTot=108.7771, largeErrors=0.13250, largeErrorsTot=0.16205, evals=  4822.9, evalsTot= 17381.9
mixed_min      : success=0.50659, error=  2.1420, errorReduced=  0.1782, errorTot=6771321335269173.0000, largeErrors=0.07977, largeErrorsTot=0.10659, evals=  1734.6, evalsTot=  2172.6
constrained_max: success=0.57091, error=4504648188.6076, errorReduced=  0.1868, errorTot=130.4789, largeErrors=0.03205, largeErrorsTot=0.04318, evals=  1220.8, evalsTot=  2703.0
binsearch      : success=0.71182, error= 14.1099, errorReduced=  0.1199, errorTot= 47.2506, largeErrors=0.01909, largeErrorsTot=0.04773, evals= 16281.7, evalsTot= 82755.7
VM             : success=0.24273, error= 10.8335, errorReduced=  0.0898, errorTot=  0.9376, largeErrors=0.00341, largeErrorsTot=0.00386, evals=  2049.4, evalsTot=  2942.8
gridsearch     : success=0.76841, error= 59.3983, errorReduced=  0.0433, errorTot=120.6592, largeErrors=0.14159, largeErrorsTot=0.17455, evals=147042.5, evalsTot=220170.5
"""],
    ["11", 1000,
"""
Wald           : success=0.03818, error=225.2044, errorReduced=  0.7528, errorTot= 37.6608, largeErrors=0.07614, largeErrorsTot=0.07614, evals=     0.0, evalsTot=     0.0
RVM            : success=0.90659, error=  3.8319, errorReduced=  0.0230, errorTot=  6.9243, largeErrors=0.00955, largeErrorsTot=0.01659, evals=  3170.7, evalsTot= 10291.7
RVM_psI        : success=0.87318, error=  3.9953, errorReduced=  0.0131, errorTot= 29.9446, largeErrors=0.01386, largeErrorsTot=0.02614, evals=  3090.3, evalsTot= 11212.8
bisection      : success=0.75773, error= 59.9321, errorReduced=  0.0340, errorTot= 51.4892, largeErrors=0.11636, largeErrorsTot=0.13545, evals=  3824.5, evalsTot= 15607.8
mixed_min      : success=0.57864, error=  3.2899, errorReduced=  0.1193, errorTot=3262584376525311.5000, largeErrors=0.05955, largeErrorsTot=0.08568, evals=  1574.6, evalsTot=  2034.5
constrained_max: success=0.63523, error=163.7362, errorReduced=  0.0899, errorTot= 66.6031, largeErrors=0.02932, largeErrorsTot=0.04068, evals=  1088.0, evalsTot=  2486.4
binsearch      : success=0.75273, error=  9.3412, errorReduced=  0.0976, errorTot= 32.5381, largeErrors=0.01091, largeErrorsTot=0.03591, evals= 15150.5, evalsTot= 73976.1
VM             : success=0.32068, error=  7.2177, errorReduced=  0.1128, errorTot=  0.4650, largeErrors=0.00386, largeErrorsTot=0.00432, evals=  1915.2, evalsTot=  3351.9
gridsearch     : success=0.80159, error= 31.3783, errorReduced=  0.0334, errorTot= 66.6911, largeErrors=0.13591, largeErrorsTot=0.15909, evals=180081.8, evalsTot=243021.1
"""],
    ["11", 3000,
"""
Wald           : success=0.08659, error=159.3901, errorReduced=  0.4562, errorTot= 17.1385, largeErrors=0.07205, largeErrorsTot=0.07205, evals=     0.0, evalsTot=     0.0
RVM            : success=0.93659, error=  3.7162, errorReduced=  0.0099, errorTot=  7.9023, largeErrors=0.00886, largeErrorsTot=0.01341, evals=  2190.8, evalsTot=  7451.9
RVM_psI        : success=0.90818, error=  3.8506, errorReduced=  0.0128, errorTot=  8.0271, largeErrors=0.00977, largeErrorsTot=0.01523, evals=  2229.1, evalsTot=  8263.6
bisection      : success=0.82432, error= 30.3407, errorReduced=  0.0485, errorTot= 16.4771, largeErrors=0.08705, largeErrorsTot=0.10091, evals=  3125.4, evalsTot= 10866.5
mixed_min      : success=0.69159, error=  2.3051, errorReduced=  0.0453, errorTot=15814584170206738.0000, largeErrors=0.02364, largeErrorsTot=0.05068, evals=  1401.5, evalsTot=  1877.6
constrained_max: success=0.72977, error= 30.9454, errorReduced=  0.0352, errorTot= 19.2405, largeErrors=0.02182, largeErrorsTot=0.02795, evals=  1051.8, evalsTot=  2059.0
binsearch      : success=0.82205, error=  4.0463, errorReduced=  0.0395, errorTot= 22.6494, largeErrors=0.00659, largeErrorsTot=0.02523, evals= 13667.2, evalsTot= 58230.7
VM             : success=0.46341, error=  2.5814, errorReduced=  0.0403, errorTot=  0.2392, largeErrors=0.00205, largeErrorsTot=0.00227, evals=  1472.1, evalsTot=  3929.4
gridsearch     : success=0.85000, error= 11.6432, errorReduced=  0.0216, errorTot= 31.7003, largeErrors=0.09636, largeErrorsTot=0.11568, evals=205155.5, evalsTot=407456.8
"""],
    ["11", 10000,
"""
Wald           : success=0.16909, error= 90.9859, errorReduced=  0.2847, errorTot=  8.7071, largeErrors=0.06205, largeErrorsTot=0.06205, evals=     0.0, evalsTot=     0.0
RVM            : success=0.98227, error=  3.1278, errorReduced=  0.0038, errorTot=  1.1732, largeErrors=0.00750, largeErrorsTot=0.00750, evals=  1229.2, evalsTot=  3773.7
RVM_psI        : success=0.95614, error=  3.2261, errorReduced=  0.0039, errorTot=  1.1757, largeErrors=0.00659, largeErrorsTot=0.00659, evals=  1101.4, evalsTot=  4441.4
bisection      : success=0.93273, error=  9.3139, errorReduced=  0.0211, errorTot=  1.6629, largeErrors=0.03886, largeErrorsTot=0.04114, evals=  2307.8, evalsTot=  6408.4
mixed_min      : success=0.81182, error=  1.5260, errorReduced=  0.0194, errorTot= 19.6726, largeErrors=0.00386, largeErrorsTot=0.02227, evals=  1300.8, evalsTot=  1550.9
constrained_max: success=0.85659, error=  9.8237, errorReduced=  0.0219, errorTot=  4.3433, largeErrors=0.01000, largeErrorsTot=0.01341, evals=  1036.5, evalsTot=  1537.3
binsearch      : success=0.91205, error=  0.9294, errorReduced=  0.0003, errorTot=  3.2472, largeErrors=0.00227, largeErrorsTot=0.00568, evals= 11049.7, evalsTot= 35509.1
VM             : success=0.69864, error= 11.3392, errorReduced=  0.1237, errorTot=  1.2125, largeErrors=0.01250, largeErrorsTot=0.01295, evals=  1181.5, evalsTot=  2684.7
gridsearch     : success=0.87932, error=  1.9279, errorReduced=  0.0044, errorTot=  6.8491, largeErrors=0.07273, largeErrorsTot=0.08568, evals= 96252.3, evalsTot=517540.5
"""],



]


dtype = [("success", float), ("error", float), 
         ("errorReduced", float), ("errorTot", float), ("largeErrors", float),
         ("largeErrorsTot", float), ("evals", float), ("evalsTot", float)]


DATA = [[*d[:2], np.genfromtxt(re.sub("[^\d\.\n\,]", "", d[2]).strip().split('\n'), 
                               delimiter=",", dtype=dtype)]
        for d in DATA]
METHODS = np.array(METHODS)
METHODNAMES = np.array(METHODNAMES)
PLOTORDER = np.array(PLOTORDER)


def get_feature(feature):
    data = np.array([d[feature] for _, _, d in DATA]).T
    print(feature, "========================")
    for fun in np.max, np.min, np.mean, np.median:
        print(fun, "----------------")
        for method, row in zip(METHODNAMES, data):
            print("{:<15} {:6.3f}".format(method, fun(row)))
    
def plot_feature(key, feature, getLegend=False):
    data = [d[1:] for d in DATA if d[0]==key]
        
    pltData = np.array([d[feature] for _, d in data]).T
    markers = ["o", "v", "s", "p", "D", ">", "*", "X", "."]
    x = [d[0] for d in data]
    figsize = (3.5, 3)
    rect = [0.15, 0.18, 0.75, 0.78]
    ax = plt.figure(figsize=figsize).add_axes(rect)
    ax.locator_params(nticks=3, nbins=3)
    order = np.argsort(PLOTORDER)
    
    if COMPARE_PSEUDOINVERSE:
        order = order[:2]
    else:
        order = np.delete(order, 1)
    
    colors = list(plt.rcParams['axes.prop_cycle'].by_key()['color'])
    tmp = colors[0]
    colors[0] = colors[1]
    colors[1] = tmp
    linestyle_cycler = cycle(["-","--","-.",":"])
    
    for method, d, c, m, l, o in zip(METHODNAMES[order], pltData[order], colors,
                                     markers, linestyle_cycler, 
                                     range(order.size)[::-1]):
        if method=="Wald" and feature=="evals":
            continue
        plt.plot(x, d, label=method, color=c, marker=m, linestyle=l,
                 alpha=0.6, zorder=o)
    plt.xscale("log")
    if feature=="error":
        plt.yscale("symlog")
        plt.ylim((-0.5,1000))
    elif feature=="evals":
        plt.yscale("log")
        plt.ylim((100,1e6))
    elif feature=="success":
        if COMPARE_PSEUDOINVERSE:
            plt.ylim((0.75,1.05))
        else:
            plt.ylim((-0.05,1.05))
        
    if getLegend:
        plt.legend(loc='lower right', fontsize=11.5, #frameon=False,
                   fancybox=True, framealpha=0.5)
        #figLegend = plt.figure(figsize = (2,2))
        #plt.figlegend(*ax.get_legend_handles_labels(), loc='upper left',
        #              fancybox=True, frameon=False)
    
    if COMPARE_PSEUDOINVERSE:
        addition = "-psi"
    else:
        addition = ""
    plt.savefig(FOLDER + feature + key + addition + ".pdf")
    #plt.show()

def createPlotsKey(key):
    plot_feature(key, "success")
    plot_feature(key, "error")
    plot_feature(key, "evals")

def createPlots():
    createPlotsKey("11")
    createPlotsKey("3")
    createPlotsKey("11cx")

if __name__ == '__main__':
    #createPlots()
    #plot_feature("11cx", "success", True)
    get_feature("error")
    get_feature("errorReduced")
    get_feature("largeErrors")
    get_feature("success")
    plt.show()


'''
Setup of the package ci_rvm
'''

from setuptools import setup

# read the contents of your README file
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

PATHADD = 'ci_rvm/'
PACKAGEADD = PATHADD.replace("/", ".")

setup(
    name="ci_rvm",
    version="0.9.0.b1",
    install_requires=['numpy', 'scipy', 'matplotlib', 'vemomoto_core_tools'], 
    packages=[PACKAGEADD[:-1]],
    package_data={
        '': ['*LICENSE*'],
    },
    license='LGPLv3',
    python_requires='>=3.6',

    # metadata to display on PyPI
    author="Samuel M. Fischer",
    description="An algorithm to find profile likelihood confidence intervals", 
    long_description=long_description,
    long_description_content_type='text/markdown',
    keywords="confidence interval, likelihood, profile likelihood", 
    url="https://github.com/vemomoto/vemomoto",
    project_urls={
        "Bug Tracker": "https://github.com/vemomoto/vemomoto/issues",
        "Documentation": "https://vemomoto.github.io/ci_rvm",
        "Source Code": "https://github.com/vemomoto/vemomoto/tree/master/ci_rvm",
        "Publication": "https://arxiv.org/abs/2004.00231",
    },
    classifiers=[
        'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
    ],
)



'''
'''
import numpy as np
import scipy.optimize as op
from scipy.stats import chi2
from scipy import linalg
import warnings
from matplotlib import pyplot as plt
import traceback
import sys
from functools import partial

try:
    from .ci_rvm import *
except ImportError:
    from ci_rvm import *


def is_negative_semidefinite(M, tol=1e-6, return_singular=False): #
    try:
        np.linalg.cholesky(-M)
        if return_singular:
            return True, False
        return True
    except np.linalg.LinAlgError:
        result = (np.linalg.eigh(M)[0] <= tol).all()
        if return_singular:
            return result, True
        return result

        
def find_profile_CI_bound(index, direction, x0, fun, jac, hess, alpha=0.95, 
                          fun0=None, *args, **kwargs):
    diff = chi2.ppf(alpha, df=1)/2
    
    if fun0 is None:
        fun0 = fun(x0)
    
    target = fun0-diff
    
    return find_CI_bound(index, target, x0, fun, jac, hess, direction==1, 
                         fun0, None, *args, track_x=True, track_f=True, 
                         **kwargs)


def find_profile_CI_bound_steps(index, x0, fun, jac, hess, direction=1, alpha=0.95, 
                          stepn=1, fun0=None, hess0=None, nmax=200, 
                          epsilon=1e-4, disp=True, vm=False):
    dim = len(x0)
    diff = chi2.ppf(alpha, df=1)/2
    
    if fun0 is None:
        fun0 = fun(x0)
    
    steps = np.linspace(fun0, fun0-diff, stepn+1)
    print("diff", diff)
    print("steps", steps) 
    
    dtype = [("params", str(dim)+'double'),
             ('logL', 'double'),
             ("x_track", object),
             ("f_track", object),
             ]
    result = np.zeros(stepn + 1, dtype=dtype)
    
    result['logL'][0] = steps[0]
    result["params"][0] = x0
    result["x_track"][0] = [x0]
    result["f_track"][0] = [fun0]
    
    for i, target in enumerate(steps[1:]):
        op_result = find_CI_bound(index, target, x0, fun, jac, hess, direction==1, 
                                  fun0, None, hess0, nmax, disp=disp, 
                                  track_x=True, track_f=True)
        x0 = op_result.x
        if not op_result.success:
            if op_result.status == 3:
                warnings.warn("Iteration limit exceeded for variable "
                              + str(index) + " and logL="
                              + str(target) + " in direction " + str(direction))
            elif op_result.status >= 4:
                warnings.warn("Optimization failed for variable "
                              + str(index) + " and logL="
                              + str(target) + " in direction " + str(direction))
                #if not np.isnan(x0).any():
                #    x0 = op_result.x
                op_result.x += np.nan
                op_result.fun += np.nan
        result["x_track"][i+1] = np.array(op_result.x_track[::direction])
        result["f_track"][i+1] = op_result.f_track[::direction]
        result["logL"][i+1] = op_result.fun
        result["params"][i+1] = op_result.x
        fun0 = op_result.fun
        hess0 = None
        
        if disp:
            try:
                print("*** logL={}; index={}, direction={}, fdiff={}; jdiff={}, nit={}".format(
                        fun0, index, direction, op_result.fun-target, np.max(np.abs(op_result.jac)), op_result.nit))
            except Exception:
                print("*** finished, but can't print due to exception.", op_result.jac)
    
    if direction < 0:
        result = result[::-1]
            
    return result
    


def find_CI_bound(
        index: "index of the parameter to consider", 
        target: "target log-likelihood l*", 
        x0: "maximum likelihood estimator (MLE)", 
        fun: "log-likelihood function", 
        jac: "gradient of fun", 
        hess: "hessian of fun", 
        forward: "True, if right end point of CI is sought, else False" = True, 
        fun0: "log-likelihood at MLE" = None, 
        jac0: "gradient of log-liekelihood at MLE" = None, 
        hess0: "Hessian of log-likelihood at MLE" = None, 
        nmax: "maximal number of iterations" = 200, 
        nchecks: "maximal number of trust-region changes per iteration" = 65,
        apprxtol: "relative tolerance between f and its approximation" = 0.5, #0.2 a 0.8 b
        resulttol: "tolerance of the result (f and norm(jac))" = 1e-3, #g
        singtol: "tolerance for singularity checks" = 1e-4, #1e-5, i
        minstep: "controls the minimal radius of the trust region" = 1e-5, 
        radiusFactor: "In [1, 2]. Controls how quickly the trust region decreases" = 1.5,
        infstep: "Stepsize after which a parameter is deemed unestimbale" = 1e10, 
        maxRadius: "radius of the trust region in the last iteration" = 1e4,
        disp: "whether to print a status message in each iteration" = True, 
        track_x: "whether to return the parameter trace" = False, 
        track_f: "whether to return the log-likelihood trace" = False):
    
    nuisance_minstep = 1e-3
    
    # ----- PREPARATION --------------------------------------------------------
    if fun0 is None:
        fun0 = fun(x0)
    if jac0 is None:
        jac0 = jac(x0)
    if hess0 is None:
        hess0 = hess(x0)
        
    # wrap the functions so that the correct end point is found
    if not forward:
        fun2 = fun
        jac2 = jac
        hess2 = hess
        flip = Flipper(index)
        fun = lambda x: fun2(flip(x)) 
        jac = lambda x: flip(jac2(flip(x)))
        hess = lambda x: flip(hess2(flip(x)))
        x0 = x0.copy()
        x0[index] *= -1
        jac0[index] *= -1
        hess0 = flip(hess0)
    else: 
        # this is just to ensure the results are traced and returned correctly
        flip = lambda x: x
    
    i: "current iteration" = 0
    x_track: "trace of parameter values" = []
    f_track: "trace of log-likelihood values" = []
    maximizing: "flag controlling whether we are maximizing f w.r.t. x[index]" = False
    nuisanceRadius: "search radius for unconstrained subproblems" = 1
    
    # counters to keep track how often functions have been called
    fun = CounterFun(fun)
    jac = CounterFun(jac)
    hess = CounterFun(hess)
    
    multi_dot = np.linalg.multi_dot
    
    is_negative_semidefinite_ = partial(is_negative_semidefinite, tol=singtol)
    
    def fPredicted_fun():
        """
        Taylor approximation of f
        """
        dx = xTmp-x
        return f + np.dot(J, dx) + multi_dot((dx, H, dx))/2
    
    def JPredicted_fun():
        """
        Taylor approximation of jac
        """
        return J + np.dot(H, xTmp-x)
    
    norm: "applied norm" = lambda arr: np.sqrt(np.mean(arr*arr))
    
    def jac_approx_precise(JPredicted_, JActual_, J_, bound=apprxtol):
        """
        Returns True, iff the predicted gradient is close to the
        actual one or, even better, closer to the target than predicted
        """
        # note how far we are from the target likelihood
        if (not searchmode=="normal" or maximizing) and not closeToDisc:
            return True
            
        return norm(JActual_) / norm(J_) <= bound or norm(JPredicted_-JActual_) / norm(J_) <= bound
    
    # TODO Accept improvements only then always, if they are better by a certain constant (compare values - line search)
    def f_approx_precise(fPredicted, fActual, bound=apprxtol):
        """
        Returns True, iff the predicted log-likelihood value is close to the
        actual one or, even better, closer to the target than predicted
        """
            
        # steps to negative predicted values cannot occur in theory and are 
        # either due to numerical issues or because we tried a very long step. 
        # Therefore, we reject these steps right away, if fActual is below the 
        # target.
        if fPredicted < target - resulttol and f > target and fActual < target:
            return False
        
        # return True if result is better than expected
        if fActual > fPredicted and xiTmp >= 0:
            return True
        
        # if we are maximizing w.r.t. the nuisance parameters, 
        if searchmode=="max_nuisance" or searchmode=="max_nuisance_const" or maximizing:
            # we do not accept steps that are worse than the current stage
            # (we guarantee elsewhere that it IS possible to increase f)
            if fActual < f: 
                if not relax_maximum_constraint or np.abs(fPredicted-fActual) > resulttol:
                    return False
            bound *= 2
            if relax_maximum_constraint:
                bound *= 2
        elif searchmode=="normal":
            # if we are outside the admissible region, we enforce that
            # we get closer to it again
            if fActual < target-resulttol and np.abs(fActual-target) > np.abs(target-f):
                return False
            elif np.abs(fActual-target) < resulttol and np.abs(fPredicted-target) < resulttol:
                return True
            
            elif fActual > f > target:
                # we relax the precision requirement, if the step is not of
                # disadvantage and brings us forward (we value progression in 
                # the paramater of interest higher than a maximal likelihood)
                bound *= 2
                
        
        # we require that the error relative to the distance from the target
        # is not larger than the given bound.
        return np.abs(fPredicted-fActual) / np.abs(target-f) <= bound
    
    def test_precision() -> "True, if appxomiations are sufficiently precise":
        """
        Tests the precision of the Taylor approximation at the suggested 
        parameter vector.
        """
        fActual = fun(xTmp)
        fPredicted = fPredicted_fun()
        
        if not f_approx_precise(fPredicted, fActual):
            return False, fActual, None
        
        if np.abs(f-target) > resulttol and not closeToDisc: 
            return True, fActual, None
        
        JActual = jac(xTmp)
        JPredicted = JPredicted_fun() 
        
        JActual_ = JActual[considered]
        JPredicted_ = JPredicted[considered]
        jac_precise = jac_approx_precise(JPredicted_, JActual_, J_)
        
        return jac_precise, fActual, JActual
        
    def is_concave_down():
        """
        Checks whether the profile likelihood function is concave down.
        """
        return a < 0
    
    if disp:
        def dispprint(*args, **kwargs):
            print(*args, **kwargs)
    else:        
        def dispprint(*args, **kwargs):
            pass
    
    # variable definitions -------------------------------------------------
    x:      "parameter vector at current iteration"      = x0
    xTmp:   "potential new parameter vector"
    xi:     "parameter of interest at current iteration" = x0[index]
    f:      "log-likelihood at current iteration"        = fun0
    J:      "gradient at current iteration"              = jac0
    H:      "Hessian at current iteration"               = hess0
    maxFun: "maximal log-likelihood"                     = fun0
    xi_max: "maximal value for the parameter of interest with f(xi, x_) >= f*"    = xi
    x_max:  "parameter vector where xi_max leads to a log-likelihood value >= f*" = x
    considered:     "Array indexing nuisance parameters" = np.ones_like(x0, dtype=bool)
    considered[index]  = False
    consideredOriginal = considered.copy()
    hessConsidered: "Array indexing nuisance Hessian"    = np.ix_(considered, considered)
    H_:     "nuisance Hessian"                           = H[hessConsidered]
    H0_:    "nuisance entries of Hessian row corresponding to the parameter of interest" = H[index][considered]
    discontinuity_counter: "steps until discontinuous parameters are released" = 0
    maximizing: "flag to memorize whether the target has been changed temporarily" = False
    resultmode: "denotes whether or what kind of result has been found" = "searching"
    radius: "radius of the trust region in the last iteration" = np.inf
    closeToDisc: "True if we ar close to a discontinuity" = False
    free_params_exist: "True if the nuisance Hessian does not have full rank" = False
    nuisance_dim: "dimension of the nuisance vector" = x.size-1
    xi_hist = []
    f_hist = []
    
    
    subproblem = FlexibleSubproblem(None, 
             lambda x: -f, 
             lambda x: -J_ - (xiTmp-xi)*H0_,
             lambda x: -H_
             )
    
    
    a = 0                        # just to prevent a possible name error...
    xiRadius = nuisanceRadius
        
    # Wrapper to catch potential errors without braking down completely
    try:    
        # ----- ITERATIONS -----------------------------------------------------
        for i in range(1, nmax+1):
            
            redoIteration = False
            ballsearch = False
            jacDiscont = False
            infstepTmp = max(infstep, 10*np.abs(x0[index]))
            #closeToDisc = False
            
            if track_x:
                x_track.append(flip(x))
            if track_f:
                f_track.append(f)
            
            # if discontinuities are found, some parameters are held constant
            # for a while. The discontinuity_counter keeps track of the time
            if discontinuity_counter:
                discontinuity_counter -= 1
            else:
                considered = consideredOriginal.copy()
                hessConsidered = np.ix_(considered, considered)
                relax_maximum_constraint = False
            
            k = -1
            
            # determine approximation
            J_ = J[considered]
            xi = x[index]
            H0 = H[index]
            H0_ = H0[considered]
            H00 = H[index, index]
            H_ = H[hessConsidered]
            
            # Determine search mode:
            # max_nuisance:  set xi to specific value, maximize other parameters
            # max_nuisance_const:  hold xi constant, maximize other parameters
            # normal:        search x with f(x) = f*, J(x) = 0
            # binary search: get back to the adissible region with a binary 
            #                search
            searchmode = "max_nuisance"
            negDefinite, singular = is_negative_semidefinite_(H_, return_singular=True)
            if negDefinite:
                if singular:
                    H_inv, rank = linalg.pinvh(H_, singtol/10, return_rank=True)
                    
                    # This should be true in genera; however, we use a different
                    # method to check for singularity here. Hence, we rather 
                    # double check that...
                    free_params_exist = rank == nuisance_dim
                else:
                    rank = nuisance_dim
                    free_params_exist = False
                    H_inv = np.linalg.inv(H_)
                    
                # in theory, the inverse of a negative definite matrix is also
                # negative definite. However, due to numerical issues this
                # may not hold here, if the matrix has unfavourable properties.
                # Therefore, we rather double check that...
                if is_negative_semidefinite_(H_inv):
                    searchmode = "normal"
                
            # ----- NORMAL SEARCH ----------------------------------------------   
            while searchmode == "normal":
                
                a = (H00 - multi_dot((H0_, H_inv, H0_)))/2
                p = J[index] - multi_dot((H0_, H_inv, J_))
                fPL0: "current value on profile log-likelihood" 
                fPL0 = f - multi_dot((J_, H_inv, J_))/2
                
                if maximizing:
                    if (is_concave_down() 
                            and not (np.allclose(a, 0, atol=singtol) and 
                                     np.allclose(((fPL0-target)/p)*(a/p), 0, 
                                                 atol=singtol))
                            ) or f < target:
                        maximizing = False
                    else:
                        targetTmp = max(fPL0 + 1, (maxFun+f)/2)
                
                if not maximizing:
                    targetTmp = target
                    
                q = fPL0 - targetTmp
                
                if a == 0:
                    sqroot = np.nan
                else:
                    sqroot = (p/a)**2 - 4*q/a 
                
                
                # if desired point is not on approximation
                if sqroot < 0:
                    if (not np.allclose(a, 0, atol=singtol) 
                            and np.linalg.matrix_rank(H, singtol) == rank): 
                            #                                )[index]):
                        if not maximizing and a<0 and f < targetTmp:
                            warnings.warn("The current function value is not on the approximation. "
                                          + "This may be due to a too large singtol. "
                                          + "The algorithm may not converge. Index=" 
                                          + str(index) + "; forward=" 
                                          + str(forward) + "; a=" 
                                          + str(a) + "; f=" + str(f)
                                          + "; root=" + str(sqroot))
                        
                        assert not maximizing
                        
                        if not is_concave_down() and f >= target:
                            # if beyond or very close to a local min
                            if p >= 0 or -p/a < singtol:
                                maximizing = True
                                continue
                            else:
                                # jump over minimum by factor 2
                                p *= 2
                            
                        sqroot = 0
                    # if f is approximately constant, we do not trust the 
                    # quadratic approximation too much and jsut try a very long
                    # step, even if it goes beyond the predicted admissible 
                    # region
                    else:
                        a = 0
                        sqroot = np.inf
                else:    
                    sqroot = np.sqrt(sqroot)
                
                # if Hessian is singular in the considered component
                # this control is important to prevent numerical errors if
                # |a| << 1 and the summands of the root are equal order of 
                # magnitude
                if (np.allclose(a, 0, atol=singtol) and 
                        np.allclose((q/p)*(a/p), 0, atol=singtol)):
                    if (maximizing and q*p>0) or (not maximizing and 
                                                    f >= target and q*p > 0):
                        maximizing = not maximizing
                        continue
                    else:
                        xiTmp = xi - q/p
                # control for NaN case
                elif p == 0 and a == 0: 
                    xiTmp = np.inf
                else:
                    xi1 = xi + (-p/a + sqroot) / 2
                    xi2 = xi + (-p/a - sqroot) / 2
                    
                    # check which root is closer to current value
                    if is_concave_down(): #not maximizing
                        assert not maximizing
                        xiTmp = max(xi1, xi2)
                    elif p*a > 0 and not maximizing and f > target:
                        maximizing = True
                        continue
                    elif maximizing and p/a < -singtol:
                        maximizing = False
                        continue
                    elif maximizing:
                        xiTmp = max(xi1, xi2)
                    else:
                        if np.abs(xi-xi1) < np.abs(xi-xi2): 
                            xiTmp = xi1
                        else:
                            xiTmp = xi2
                
                # if likelihood is independent of parameter
                if np.abs(xiTmp-xi) >= infstep * (1-singtol):
                    
                    
                    # try a very large xi value
                    xiTmp = xi + infstepTmp
                    xTmp = x.copy()
                    xTmp[index] = xiTmp
                    xTmp[considered] += -np.dot(H_inv, H0_*(xiTmp-xi)+J_)
                    fActual = fun(xTmp)
                    
                    # if still larger than target
                    if fActual >= target:
                        JActual = jac(xTmp)
                        resultmode = "unbounded"
                        break 
                    
                    if not f > target:
                        if norm(J_) > singtol:
                            searchmode = "max_nuisance"
                            break 
                        
                        searchmode = "binary_search"
                        break 
                    xiTmp = xiTmp/4 + 3*xi/4
                
                xTmp = x.copy()
                xTmp[index] = xiTmp
                xTmp[considered] += -np.dot(H_inv, H0_*(xiTmp-xi)+J_)
                
                xx = x[considered]
                upperRadius = np.linalg.norm(xx-xTmp[considered])
                
                # Test precision
                stop, fActual, JActual = test_precision()
                if stop:
                    radius = max(radius, upperRadius)
                    break
                
                if radius >= upperRadius:
                    lowerRadius = upperRadius/2
                else:
                    lowerRadius = radius
                
                if upperRadius > 0:
                    radius = upperRadius
                
                tmpRadius = lowerRadius
                xiStep = xiTmp-xi
                
                if tmpRadius == 0:
                    xiTmp = xi + min(xiStep/2, infstep)
                else:
                    xiTmp = xi + xiStep*np.power(tmpRadius/radius, np.log(2)/np.log(radiusFactor))
                
                ballsearch = True
                break
                    
            if not resultmode=="searching":
                x = xTmp
                J = JActual
                JActual_ = JActual[considered]
                f = fActual
                break
            
            if searchmode == "normal" and free_params_exist:
                J_new = norm(JPredicted_fun()[considered])
                J_current = norm(J[considered])
                
                # if ignoring the free parameters hinders us from making an 
                # improvement w.r.t. the gradient
                abstol = max(np.log2(np.abs(xiTmp-xi))*np.abs(f-target)
                             /np.abs(maxFun-target), resulttol)
                #if (J_new > abstol and J_new/J_current > apprxtol):
                if (J_new > abstol and J_new/J_current > apprxtol):
                    searchmode = "max_nuisance"            
            
            counter = 0
            repeat = True
            while repeat:
                repeat = False
                
                if counter > 10:
                    raise RuntimeError("Infinite loop", f, J, H, considered)
                
                # ----- CONSTRAINED MAXIMIZATION -----------------------------------
                if searchmode == "max_nuisance" or searchmode == "max_nuisance_const":
                    xiTmp = xi
                    tmpRadius = nuisanceRadius * np.sqrt(J_.size)
                    ballsearch = True
                    
                    if not searchmode == "max_nuisance_const":
                        if f > target:
                            xiTmp += xiRadius
                        else:
                            searchmode = "max_nuisance_const"
                    
                # ----- BALL SEARCH ------------------------------------------------   
                if ballsearch:
                    increaseTrustRegion = True
                    xiPrev = xi
                    nAdjustement = 5
                    for k in range(nchecks):
                        
                        # it can happen that the radius is so conservative that
                        # the nuisance parameters cannot be maximized 
                        # sufficiently far to keep f in the admissible region
                        nRatioAdjustement = 10
                        for l in range(nRatioAdjustement):
                            #print("tmpRadius", tmpRadius)
                            xTmp_diff, _ = subproblem.solve(tmpRadius, 
                                                            searchmode=="normal",
                                                            jac0tol=singtol)
                            if xi > xiTmp and f>target:
                                print("a, p", a, p)
                                print("q, fPL0", q, fPL0)
                                print("targetTmp", targetTmp)
                                print("xiTmp, xi", xiTmp, xi)
                            xTmp = x.copy()
                            xTmp[index] = xiTmp
                            xTmp[considered] += xTmp_diff
                            fPredicted = fPredicted_fun()
                            
                            # if we maximize w.r.t. the nuisance parameters,
                            # we want to assure that f gets increased to get
                            # back to the likelihood ridge
                            if (not searchmode == "max_nuisance" and 
                                    not searchmode == "max_nuisance_const"):
                                break
                            precise, fActual, JActual = test_precision()
                            if fPredicted >= f:
                                break
                            # if the step is super close and f still does not 
                            # increase, there must be a numerical issue. We
                            # leave the treatment of this issue to the procedures
                            # below.
                            elif (np.isclose(xiTmp, xi, atol=minstep/100) and tmpRadius<minstep/100):
                                break
                            
                            if l == nRatioAdjustement-2:
                                xiPrev = xiTmp
                                xiTmp = xi
                            else:
                                # we either increase the search radius for the
                                # nuisance parameters or decrease the search
                                # radius for the parameter of interest dependent
                                # on whether the approximation is precise and
                                # we can increase the trust region or not
                                if increaseTrustRegion and precise and tmpRadius < maxRadius*np.sqrt(xTmp_diff.size):
                                    tmpRadius *= 2
                                else:
                                    # if the error occurs even though xi is 
                                    # constant, there must be an error in the 
                                    # approximate solution and we decrease the
                                    # search radius
                                    if np.isclose(xiTmp, xi, atol=minstep/100):
                                        tmpRadius /= 5
                                    xiTmp = (xi+xiTmp) / 2
                                    
                        # if maximizing w.r.t. the nuisance parameters,
                        # adjust the trust region
                        if (searchmode == "max_nuisance" or 
                                searchmode == "max_nuisance_const"):
                            if precise:
                                # update trust region
                                nuisanceRadius = tmpRadius / np.sqrt(J_.size)
                                # update xiRadius only if we have changed xi
                                if searchmode == "max_nuisance":
                                    if xi != xiTmp:
                                        xiRadius = max(xiTmp-xi, minstep)
                                    else:
                                        # if we had to set xiRadius to 0, we set 
                                        # it to the last non-zero value
                                        xiRadius = max(xiPrev-xi, minstep)
                                    # since the xiRadius cannot recover on its
                                    # own, we always gradually increase it
                                    if 2*xiRadius < nuisanceRadius:
                                        xiRadius *= 2
                                
                                # if we have not decreased the trust region 
                                # before, try to increase it
                                if increaseTrustRegion and k < nAdjustement:
                                    xTmp2 = xTmp
                                    fActual2 = fActual
                                    xiTmp = xi + 2*xiRadius
                                    tmpRadius *= 2
                                    tmpRadius = min(maxRadius*np.sqrt(xTmp_diff.size),
                                                    tmpRadius)
                                    continue
                                else:
                                    stop = True
                            else:
                                stop = False
                                if increaseTrustRegion: 
                                    if k == 0:  
                                        # if in the first iteration, note that 
                                        # we are decreasing the trust region
                                        increaseTrustRegion = False
                                    else:
                                        # set xTmp to the last precise value
                                        fActual = fActual2
                                        xTmp = xTmp2
                                        xiTmp = xTmp[index]
                                        stop = True
                        else:    
                            stop, fActual, JActual = test_precision()
                        
                        if searchmode == "normal":
                            
                            if stop:
                                # if the approximation is precise, we may increase
                                # the radius, but save the largest current result
                                # to fall back to
                                xTmp2 = xTmp
                                fActual2 = fActual
                                lowerRadius = tmpRadius
                            else:
                                upperRadius = tmpRadius
                                
                            # if we have to decrease the trust region
                            if upperRadius <= lowerRadius and not stop:
                                # continue with the continuity checks and
                                # cut the trust trust region by half
                                pass
                            # if precision of geometric binary search is as 
                            # desired (first equality necessary to prevent NaNs)
                            elif upperRadius==0 or upperRadius/lowerRadius <= 2:
                                # stop iteration and set xTmp to the optimal 
                                # value
                                if lowerRadius > 0:
                                    radius = lowerRadius
                                fActual = fActual2
                                xTmp = xTmp2
                                xiTmp = xTmp[index]
                                break
                            else:
                                # update xiTmp and radius
                                tmpRadius = np.sqrt(lowerRadius*upperRadius)
                                xiTmp = xi + xiStep*np.power(tmpRadius/radius, np.log(2)/np.log(radiusFactor))
                                continue
                        elif stop:
                            break
                        
                        minstepTmp = max(minstep / max(norm(J_), 1), 1e-12)
                        #if (not searchmode=="normal" or maximizing) and not closeToDisc:
                        #    minstepTmp *= 1000
                        
                        # if the step is too small
                        diffs = xTmp - x
                        if not stop and np.allclose(xTmp, x, atol=minstepTmp, rtol=1e-14):
                            
                            # if we are not really on a discontinuity but only 
                            # on a local maximum with a non-negative definite
                            # Hessian, we may just release the constraint
                            # that we need to get to a larger function value
                            if (not searchmode=="normal" or maximizing) and not closeToDisc: 
                                relax_maximum_constraint = True
                                discontinuity_counter = 20
                                if test_precision()[0]:
                                    break
                                
                            # check parameter of interest first
                            xTmp = x.copy()
                            xTmp[index] += diffs[index]
                            
                            # if small change in parameter of interest makes 
                            # approximation imprecise
                            if diffs[index] and not test_precision()[0]:
                                
                                # make sure discontinuity happens on 
                                # profile function
                                if norm(J[considered]) > resulttol:
                                    relax_maximum_constraint = False
                                    searchmode = "max_nuisance_const"
                                    closeToDisc = True
                                    xiDisc = xi
                                    repeat = True
                                    radius = tmpRadius
                                    break
                                
                                fActual = fun(xTmp)
                                fCloseTarget = (np.allclose(target, f, 
                                                            atol=resulttol) and
                                                not np.allclose(target, fActual, 
                                                                atol=resulttol))
                                
                                # if step is favourable or benign, we accept 
                                # depending on the distance measure this case
                                # will never occur, as favourable steps may 
                                # always be classified as precise
                                if fActual >= target or f < fActual:
                                    pass
                                # if the target is on the jump
                                elif ((f > target or fCloseTarget)
                                      and fActual < target):
                                    # stop here
                                    fPredicted = fPredicted_fun()
                                    resultmode = "discontinuous"
                                    xTmp = x
                                    JActual = J
                                else:
                                    # get back to the feasible region by 
                                    # conducting a binary search
                                    searchmode = "binary_search"
                                    
                                dispprint("iter {:3d}{}: ***discontinuity of {} in x_{} at {}***".format(
                                            i, ">" if forward else "<", f-fActual, 
                                            index, x))
                                break
                            
                            # check for discontinuities in nuisance parameters
                            discontinuities = np.zeros_like(considered)
                            discontinuities_rejected = discontinuities.copy()
                            if not f_approx_precise(fPredicted, fActual):
                                if diffs[index]:
                                    fPredicted = fPredicted_fun()
                                    fActual = fun(xTmp)
                                else:
                                    fPredicted = fActual = f
                                for l in np.nonzero(considered)[0]:
                                    xTmp[l] += diffs[l]
                                    fActual_previous = fActual
                                    fActual = fun(xTmp)
                                    fPredicted = fPredicted_fun()
                                    if not f_approx_precise(fPredicted, fActual):
                                        discontinuities[l] = True
                                        if fActual < fActual_previous:
                                            discontinuities_rejected[l] = True
                                            xTmp[l] = x[l]
                                            fActual = fActual_previous
                                jacDiscont = False
                            # if the gradient is responsible for the rejection
                            # we require a smaller steap
                            elif np.abs(f-target) < resulttol and np.allclose(xTmp, x, rtol=minstepTmp/10):
                                for l in np.nonzero(considered)[0]:
                                    xTmp[l] += diffs[l]
                                    JActual_previous = JActual
                                    JActual = jac(xTmp) 
                                    JPredicted = JPredicted_fun()
                                    jac_precise = jac_approx_precise(JPredicted[considered], JActual[considered], J[considered]) 
                                    if not jac_precise:
                                        discontinuities[l] = True
                                        if not (JActual_previous[l] * J[l] >= 0 
                                                or diffs[l]*J[l] <= 0):
                                            discontinuities_rejected[l] = True
                                            xTmp[l] = x[l]
                                            JActual = JActual_previous
                                jacDiscont = True
                            

                            if discontinuities.any():
                                
                                varStr = "variables" if not jacDiscont else "gradient entries"
                                    
                                dispprint("iter {:3d}{}: ***index={}: discontinuities in {} {} at {}***".format(
                                            i, ">" if forward else "<", index, varStr,
                                            tuple(np.nonzero(discontinuities)[0]), x))
                                
                                redoIteration = discontinuities.sum() == discontinuities_rejected.sum()
                                
                                if discontinuities_rejected.any():
                                    discontinuity_counter = 25
                                    considered &= ~discontinuities_rejected
                                    hessConsidered = np.ix_(considered, considered)
                                    relax_maximum_constraint = False
                                
                                break
                        if tmpRadius > (radiusFactor**nchecks-k)*minstep:
                            tmpRadius /= 5
                            xiTmp = (4*xi+xiTmp) / 5
                        else:
                            tmpRadius /= radiusFactor
                            xiTmp = (xi+xiTmp) / 2
                        
                    else:
                        xTmp = x
                        JActual = J
                        fActual = f
            
            if redoIteration:
                continue
                    
            # ----- BINARY SEARCH ----------------------------------------------
            if searchmode == "binary_search":
                tol = 0.01
                bin_min = x
                bin_max = x_max
                for k in range(nchecks):
                    xTmp = (bin_max+bin_min) / 2
                    fActual = fun(xTmp)
                    
                    if 0 <= fActual - target < tol:
                        break
                    
                    if fActual < target:
                        bin_min = xTmp
                    else:
                        bin_max = xTmp
                else: 
                    xTmp = bin_max
                JActual = jac(xTmp)
                        
                        
            # ----- UPDATES & CEHCKS -------------------------------------------
            if (xTmp == x).all() and not jacDiscont and resultmode=="searching":
                dispprint("iter {:3d}{}: ***no improvement when optimizing x_{} at {}***".format(
                            i, ">" if forward else "<", index, flip(x)))
                resultmode = "discontinuous"
            
            
            
            xiChange = xTmp[index] - xi
            xChange = norm(x-xTmp)
            
            # when we are maximizing the nuisance parameters, we may step 
            # farther than the allowedbound
            if xiChange >= infstep and f >= target:
                resultmode = "unbounded"
            elif (searchmode == "max_nuisance" or searchmode == "max_nuisance_const"
                  or maximizing) and (xiChange < nuisance_minstep 
                                      and f > target+resulttol):
                if (len(xi_hist) > 5 and (np.abs(np.array(xi_hist[-5:]) - xi) < nuisance_minstep).all() and
                    (np.abs(np.array(f_hist[-5:]) - f) < nuisance_minstep).all()):
                    
                    xTmp[index] += nuisance_minstep
                    fActual = fun(xTmp)
                    m = 0
                    fActual_previous = fActual
                    xTmp_previous = xTmp.copy()
                    while fActual > target and m < 1000:
                        fActual_previous = fActual
                        xTmp_previous = xTmp.copy()
                        xTmp[index] += nuisance_minstep * 2**m
                        fActual = fun(xTmp)
                        m += 1
                    
                    xTmp = xTmp_previous
                    fActual = fActual_previous
                    
                    m = -1
                    while fActual < target-50 and m > -10:
                        fActual_previous = fActual
                        xTmp_previous = xTmp.copy()
                        xTmp[index] += nuisance_minstep * 2**m
                        fActual = fun(xTmp)
                        m -= 1
                    
                    
                    JActual = jac(xTmp)
                    
                    dispprint("iter {:3d}{}: ***taking additional step in x_{} to avoid convergence issues. f={:6.3}***".format(
                        i, ">" if forward else "<", index, fActual))
                    for m in range(1, 5):
                        xi_hist[-m] += xTmp[index]-xi
                xi_hist.append(xTmp[index])
                f_hist.append(fActual)
            
            x = xTmp
            xi = xTmp[index]
            
            if JActual is None:
                JActual = jac(xTmp)
            JImprovement = np.log2(norm(J[considered])/norm(JActual[considered]))
            fImprovement = np.log2(np.abs((f-target)/(fActual-target)))
            
            J = JActual
            f = fActual
            JActual_ = JActual[considered]
            
            # reset maximal value
            if f > target and xi > xi_max:
                xi_max = xi
                x_max = x
            
            fPredicted = fPredicted_fun()
            
            if disp:
                radius_str = ""
                if searchmode == "normal":
                    if maximizing:
                        mode = "maximizing"
                    else:
                        mode = "searching"
                elif (searchmode == "max_nuisance" 
                      or searchmode == "max_nuisance_const"):
                    mode = "maximizing nuisance parameters"
                    radius_str = "; step={}; radius={}".format(
                                        __round_str(xiRadius), __round_str(nuisanceRadius))
                elif searchmode == "binary_search":
                    mode = "binary search"
                else:
                    mode = "other"
                
                ndiscont = np.sum(~considered)-1
                if ndiscont:
                    discontStr = "; ndisct={}".format(ndiscont)
                else: 
                    discontStr = ""
                    
                dispprint(("iter {:3d}{}: x_{}_d={}; f_d={}; jac_d={}; " + 
                       "nsteps={:2d}; x_d={}; f_impr={}; jac_impr={}; " +
                       "f_e={}{}{} - {}").format(i, 
                                                 ">" if forward else "<", index, 
                                                 __round_str(xi-x0[index]), 
                                                 __round_str(f-target), 
                                                 __round_str(norm(JActual_)), k+2, 
                                                 __round_str(xChange), 
                                                 __round_str(fImprovement), 
                                                 __round_str(JImprovement), 
                                                 __round_str(fActual-fPredicted), 
                                                 discontStr, radius_str, mode))
            
            
            if resultmode == "searching":
                resViol = max(norm(JActual_), np.abs(f-target))
                if resViol <= resulttol:
                    resultmode = "success"
                    break
            else:
                break
            
            if f > maxFun:
                dispprint("-> iter {:3d}{}: !!!found NEW MAXIMUM for x_{} of {:6.3f} (+{:6.3f}) at {}!!!".format(
                        i, ">" if forward else "<", index, f, f-fun0, flip(x)))
                maxFun = f
            
            if closeToDisc:
                if np.abs(xi - xiDisc) > minstep:
                    closeToDisc = False
            
            H = hess(x)
        
        else:
            resultmode = "exceeded"
    except np.linalg.LinAlgError:
        resultmode = "linalg_error"
        exc_type, exc_value, exc_traceback = sys.exc_info()
        traceback.print_exception(exc_type, exc_value, exc_traceback,
                                  file=sys.stdout)
    except Exception:
        resultmode = "error"
        exc_type, exc_value, exc_traceback = sys.exc_info()
        traceback.print_exception(exc_type, exc_value, exc_traceback,
                                  file=sys.stdout)

    
    if resultmode == "success":
        status = 0
        success = True
    elif resultmode == "discontinuous":
        status = 1
        success = False
    elif resultmode == "unbounded":
        status = 2
        success = True
    elif resultmode == "exceeded":
        success = False
        status = 3
    elif resultmode == "error" or resultmode == "linalg_error":
        #x = x0+np.nan
        #f = np.nan
        #JActual_ = x0[considered] + np.nan
        try:        
            JActual_ = J_
        except UnboundLocalError:
            JActual_ = jac(x)[considered]
        success=False
        if resultmode == "error":
            status = 5
        else:
            status = 4
    
    dispprint(op.OptimizeResult(x=flip(x), 
                         fun=f,
                         jac=JActual_,
                         success=success, status=status,
                         nfev=fun.evals, njev=jac.evals, nhev=hess.evals, 
                         nit=i,
                         message=STATUS_MESSAGES[status]
                         ))
    
    if track_f:
        f_track.append(f)
    if track_x:
        x_track.append(flip(x))
    
    return op.OptimizeResult(x=flip(x), 
                             fun=f,
                             jac=JActual_,
                             success=success, status=status,
                             nfev=fun.evals, njev=jac.evals, nhev=hess.evals, 
                             nit=i,
                             x_track=np.array(x_track),
                             f_track=np.array(f_track),
                             message=STATUS_MESSAGES[status]
                             )
    
